"""Test the hkl.util module."""

import json
import time

import databroker
import pytest
from bluesky import plans as bp
from bluesky.run_engine import RunEngine
from packaging import version

from .. import util
from .tools import DocsCollector
from .tools import safe_json_dump
from .tools import validate_descriptor_doc_content

NO_SUCH_PACKAGE_NAME = "no-such-package"


@pytest.fixture(scope="function")
def cat():
    yield databroker.temp().v2


@pytest.fixture(scope="function")
def RE(cat):
    import bluesky

    engine = bluesky.RunEngine()
    engine.subscribe(cat.v1.insert)
    yield engine


@pytest.fixture(scope="function")
def fourc():
    """4-circle with renamed axes and oriented sample."""
    from ophyd import Component
    from ophyd import SoftPositioner

    from hkl import E4CV
    from hkl import SimMixin

    class FourCircle(SimMixin, E4CV):
        theta = Component(SoftPositioner, kind="hinted", init_pos=0)
        chi = Component(SoftPositioner, kind="hinted", init_pos=0)
        phi = Component(SoftPositioner, kind="hinted", init_pos=0)
        ttheta = Component(SoftPositioner, kind="hinted", init_pos=0)

    fourc = FourCircle("", name="fourc")
    # rename the physical axes
    fourc.calc.physical_axis_names = {
        # E4CV: local
        "omega": "theta",
        "chi": "chi",
        "phi": "phi",
        "tth": "ttheta",
    }

    # fourc.wait_for_connection()
    fourc._update_calc_energy()
    crystal_setup(fourc)

    yield fourc


def crystal_setup(diffractometer):
    from hkl import Lattice

    diffractometer.calc.wavelength = 1.0
    a0 = 5.4321
    # fmt: off
    diffractometer.calc.new_sample(
        "vibranium",
        lattice=Lattice(a=a0, b=a0, c=a0, alpha=90, beta=90, gamma=90)
    )

    diffractometer.calc.sample.add_reflection(
        1, 2, 3,
        position=diffractometer.calc.Position(
            ttheta=60,
            theta=40,
            chi=0,
            phi=0,
        ),
    )
    # fmt: on


def test__package_info_states():
    assert util._package_info is None
    util.get_package_info("hkl")
    assert util._package_info is not None


@pytest.mark.parametrize(
    "package_name, minimum_version",
    [
        ("bluesky", "1.6"),
        ("pygobject", "3.40"),
        ("hkl", "5.0.0"),
        ("hklpy", "0"),  # minimum test for unversioned use
        (NO_SUCH_PACKAGE_NAME, "---"),
        # ("ophyd", "1.6"),  conda has right version but pip has 0.0.0
    ],
)
def test_get_package_info(package_name, minimum_version):
    v = util.get_package_info(package_name)
    if v is None:
        assert package_name in ("hklpy", NO_SUCH_PACKAGE_NAME)
    else:
        assert "version" in v
        v_string = v.get("version", "unknown")

        if package_name == "hkl":
            assert v_string.startswith("5.0.0.")

        v_package = version.parse(v_string)
        assert v_package >= version.parse(minimum_version)


@pytest.mark.parametrize(
    # fmt: off
    "case",
    [
        (None),
        ([]),
        ("")
    ],
    # fmt: on
)
def test_software_versions_default_list(case):
    v = util.software_versions(case)
    assert isinstance(v, dict)
    expected = sorted(util.DEFAULT_PACKAGE_LIST)
    assert sorted(v.keys()) == expected


@pytest.mark.parametrize(
    "package_name, minimum_version",
    [
        ("bluesky", "1.6"),
        ("pygobject", "3.40"),
        ("hkl", "5.0.0"),
        ("hklpy", "0"),  # minimum test for unversioned use
        (NO_SUCH_PACKAGE_NAME, "---"),
        # ("ophyd", "1.6"),  conda has right version but pip has 0.0.0
    ],
)
def test_software_versions_items(package_name, minimum_version):
    v = util.software_versions([package_name])
    if package_name in v:
        v_string = v[package_name]
        v_package = version.parse(v_string)
        assert v_package >= version.parse(minimum_version)
    else:
        assert package_name in ("hklpy", NO_SUCH_PACKAGE_NAME)


def test_issue215(cat, RE, fourc):
    """restore_reflections(orientation, fourc) cannot find renamed positioner."""
    from bluesky import plans as bp

    canonical_names = "omega chi phi tth".split()
    our_names = "theta chi phi ttheta".split()
    assert our_names != canonical_names
    assert fourc.calc.physical_axis_names != canonical_names
    assert fourc.calc.physical_axis_names == our_names

    problem_reflection_dict = {
        "reflection": {"h": 1.0, "k": -3.0, "l": -1.0},
        "flag": 1,
        "wavelength": 1.1169734383241103,
        "position": {
            "omega": -8.208399999999983,
            "chi": -47.47651999999994,
            "phi": -5.684341886080802e-14,
            "tth": 25.09473789610388,
        },
        "orientation_reflection": False,
    }
    problem_positioner_names = list(problem_reflection_dict["position"].keys())
    assert problem_positioner_names == canonical_names

    assert len(cat) == 0
    uids = RE(bp.count([fourc]))
    time.sleep(1)
    assert len(uids) == 1
    assert len(cat) == 1

    orientation = util.run_orientation_info(cat[-1])
    assert isinstance(orientation, dict)
    assert fourc.name in orientation, f"{list(orientation)=!r}"
    # assert list(orientation.keys()) == []

    orient = orientation[fourc.name]
    reals = orient.get("_reals")
    assert reals is not None, f"{orient=!r}"
    assert reals != canonical_names
    assert reals == our_names
    for reflection in orient["reflections_details"]:
        assert list(reflection["position"].keys()) == canonical_names

    success = False
    try:
        # since it is just a problem of motor names in reflections ...
        util.restore_reflections(orient, fourc)
        success = True
    finally:
        assert success, "Could not restore orientation reflections."


@pytest.mark.parametrize("use_refl", [None, "short", "long"])
def test_RE_documents(use_refl, e4cv):
    """Issue #316."""

    docs = DocsCollector()  # capture the raw document stream directly
    assert len(docs.raw) == 0

    # add a reflection?
    if use_refl == "short":
        e4cv.calc.sample.add_reflection(1, 2, 3, (40, 0, 0, 60))
    elif use_refl == "long":
        # fmt: off
        e4cv.calc.sample.add_reflection(
            1, 2, 3,
            position=e4cv.calc.Position(tth=60, omega=40, chi=0, phi=0),
        )
        # fmt: on
    refs = e4cv.reflections.get()
    assert isinstance(refs, list)
    assert len(refs) == len(e4cv.calc.sample.reflections)
    if len(refs) > 0:
        assert list(refs[0]) == [1.0, 2.0, 3.0]

    safe = safe_json_dump(refs)
    assert isinstance(safe, bytes)

    assert isinstance(safe_json_dump(e4cv.reflections_details.get()), bytes)
    # assert f"{refs=!r}" == "debug"

    RE = RunEngine()
    cat = databroker.temp().v2
    RE.subscribe(cat.v1.insert)

    uids = RE(bp.count([e4cv]), docs.receiver)
    assert len(uids) == 1
    assert len(docs.raw) == 4

    # Get the configuration from the raw document stream
    validate_descriptor_doc_content(e4cv.name, docs.raw.get("descriptor"))

    # Get the configuration from the safe_json version of the document stream
    # difference: np.ndarray --> list
    # difference: tuple --> list
    descriptor = docs.safe_json.get("descriptor")
    assert isinstance(descriptor, bytes)
    validate_descriptor_doc_content(e4cv.name, json.loads(docs.safe_json.get("descriptor")))

    # Get the configuration from the databroker document stream
    for key, doc in cat.v1[uids[0]].documents():
        if key == "descriptor":
            validate_descriptor_doc_content(e4cv.name, doc)

    orientation = util.run_orientation_info(cat[uids[0]])
    assert isinstance(orientation, dict)
    assert e4cv.name in orientation



"""
Tests for changes due to issues #307 & #308.
"""

from collections import namedtuple
from contextlib import nullcontext as does_not_raise

import pytest

from ..util import get_position_tuple


@pytest.mark.parametrize(
    "miller, context, message",
    [
        [(1, 2, 3), does_not_raise(), None],
        [(1.0, 2.0, 3.0), does_not_raise(), None],
        [("1", 2, 3), pytest.raises(TypeError), "Must be number, not str"],
        [(1, 2, "3"), pytest.raises(TypeError), "Must be number, not str"],
        [([1], 2, 3), pytest.raises(TypeError), "Must be number, not list"],
        [(object, 2, 3), pytest.raises(TypeError), "Must be number, not type"],
        [(None, 2, 3), pytest.raises(TypeError), "does not allow None as a value"],
        [None, pytest.raises(TypeError), "argument after * must be an iterable, not NoneType"],
        [((1,), 2, 3), pytest.raises(TypeError), "Must be number, not tuple"],
        [
            # Tests that h, k, l was omitted, only a position was supplied.
            # This is one of the problems reported.
            namedtuple("PosAnything", "a b c d".split())(1, 2, 3, 4),
            pytest.raises(TypeError),
            "Expected positions, received 4",
        ],
    ],
)
def test_miller_args(miller, context, message, e4cv):
    """Test the Miller indices arguments: h, k, l."""
    sample = e4cv.calc.sample
    assert sample is not None

    with context as info:
        sample.add_reflection(*miller)
    if message is not None:
        assert message in str(info.value)


# fmt: off
@pytest.mark.parametrize(
    "miller, angles, context, message",
    [
        # None
        [(-1, -2, -3), None, does_not_raise(), None],
        # tuple
        [(-1, -2, -3), (1, 2, 3, 4), does_not_raise(), None],
        # list
        [(-1, -2, -3), [1, 2, 3, 4], does_not_raise(), None],
        # dict
        [
            (-1, -2, -3),
            {"omega": 1, "chi": 2, "phi": 3, "tth": 4},
            pytest.raises(TypeError),
            "Expected list, tuple, or calc.Position() object,",
        ],
        # namedtuple
        [(-1, -2, -3), "namedtuple 1 2 3 4", does_not_raise(), None],
        # calc.Position object
        [(-1, -2, -3), "Position 1 2 3 4", does_not_raise(), None],
        # Unacceptable namedtuple (does not start with "Pos")
        [
            (-10, -2, -3),
            namedtuple("Not_Position", "a b c d".split())(1, 2, 3, 4),
            pytest.raises(TypeError),
            "Expected list, tuple, or calc.Position() object",
        ],
        # Acceptable namedtuple (starts with "Pos") type
        # yet the axis names are wrong.
        [
            (-10, -2, -3),
            namedtuple("Positron", "a b c d".split())(1, 2, 3, 4),
            pytest.raises(KeyError),
            "Wrong axes names.  Expected [",
        ],
        # Acceptable namedtuple (starts with "Pos") type
        # yet values must all be numeric.
        [
            (-10, -2, -3),
            namedtuple("Post", "omega chi phi tth".split())(1, "2", 3, 4),
            pytest.raises(TypeError),
            "All values must be numeric",
        ],
        # Acceptable namedtuple (starts with "Pos") type & content.
        [
            (-10, -2, -3),
            namedtuple("Posh", "omega chi phi tth".split())(1, 2, 3, 4),
            does_not_raise(),
            None,
        ],
    ]
)
# fmt: on
def test_position_args(miller, angles, context, message, e4cv):
    """Test sample.add_reflection with the different ways positions can be entered."""
    assert len(miller) == 3

    calc = e4cv.calc
    assert calc is not None

    sample = calc.sample
    assert sample is not None

    if isinstance(angles, str):
        # These constructs require some additional setup.
        if angles.startswith("Position"):
            angles = calc.Position(*map(float, angles.split()[1:]))
        elif angles.startswith("namedtuple"):
            # fmt: off
            angles = get_position_tuple("omega chi phi tth".split())(
                *list(map(float, angles.split()[1:]))
            )
            # fmt: on

    with context as info:
        sample.add_reflection(*miller, angles)
    if message is not None:
        assert message in str(info.value)


@pytest.mark.parametrize(
    "args, context, message",
    [
        # wrong number of reals, tuple
        [[1, 2, 3, (4, 5)], pytest.raises(ValueError), "Expected 4 positions,"],
        # wrong number of reals, list
        [[1, 2, 3, [4, 5]], pytest.raises(ValueError), "Expected 4 positions,"],
        # wrong number of reals, namedtuple
        [
            [1, 2, 3, namedtuple("Position", "omega chi phi".split())(1, 2, 3)],
            pytest.raises(KeyError),
            "Wrong axes names.  Expected [",
        ],
        # wrong representation of position
        [[1, 2, 3, 4], pytest.raises(TypeError), "Expected positions"],
        # pseudos provided as tuple
        [[(1, 2, 3), (4, 5, 6, 7)], pytest.raises(TypeError), "missing 1 required"],
        # only pseudos (uses current positions so no problem)
        [[1, 2, 3], does_not_raise(), None],
        [[1, 2, 3, namedtuple("Position", "omega chi phi tth".split())(4, 5, 6, 7)], does_not_raise(), None],
    ],
)
def test_add_reflection(args, context, message, e4cv):
    calc = e4cv.calc
    sample = calc.sample
    assert sample is not None

    with context as info:
        sample.add_reflection(*args)
    if message is not None:
        assert message in str(info.value)


@pytest.mark.parametrize(
    "dname, mode, n_r, n_w, e_a_c, c_a_c",
    [
        # all modes on each common geometry, some with canonical names, some with renames
        # If not renamed, set c_a_c to `None`
        # fmt: off
        ["e4cv", "bissector", 4, 4, "", None],
        ["e4cv", "constant_chi", 4, 3, "chi", None],
        ["e4cv", "constant_omega", 4, 3, "omega", None],
        ["e4cv", "constant_phi", 4, 3, "phi", None],

        ["e4cv_renamed", "bissector", 4, 4, "", ""],
        ["e4cv_renamed", "constant_chi", 4, 3, "chi", "chi"],
        ["e4cv_renamed", "constant_omega", 4, 3, "omega", "theta"],
        ["e4cv_renamed", "constant_phi", 4, 3, "phi", "phi"],
        ["e4cv_renamed", "double_diffraction", 4, 4, "", ""],

        ["e6c", "bissector_horizontal", 6, 5, "delta", None],
        ["e6c", "bissector_vertical", 6, 4, "mu gamma", None],
        ["e6c", "constant_phi_vertical", 6, 3, "mu phi gamma", None],
        ["e6c", "psi_constant_horizontal", 6, 4, "mu delta", None],
        ["e6c", "psi_constant_vertical", 6, 4, "mu gamma", None],

        ["tardis", "bissector_horizontal", 6, 5, "delta", "gamma"],
        ["tardis", "bissector_vertical", 6, 4, "mu gamma", "theta delta"],
        ["tardis", "constant_phi_vertical", 6, 3, "mu phi gamma", "theta phi delta"],
        ["tardis", "psi_constant_horizontal", 6, 4, "mu delta", "theta gamma"],
        ["tardis", "psi_constant_vertical", 6, 4, "mu gamma", "theta delta"],
        ["tardis", "constant_chi_vertical", 6, 3, "mu chi gamma", "theta chi delta"],
        ["tardis", "constant_mu_horizontal", 6, 3, "mu omega delta", "theta omega gamma"],
        ["tardis", "constant_omega_vertical", 6, 3, "mu omega gamma", "theta omega delta"],
        ["tardis", "double_diffraction_horizontal", 6, 4, "omega delta", "omega gamma"],
        ["tardis", "double_diffraction_vertical", 6, 4, "mu gamma", "theta delta"],
        ["tardis", "lifting_detector_mu", 6, 3, "omega chi phi", "omega chi phi"],
        ["tardis", "lifting_detector_omega", 6, 3, "mu chi phi", "theta chi phi"],
        ["tardis", "lifting_detector_phi", 6, 3, "mu omega chi", "theta omega chi"],

        ["k4cv", "bissector", 4, 4, "", None],
        ["k4cv", "constant_chi", 4, 4, "", None],
        ["k4cv", "constant_omega", 4, 4, "", None],
        ["k4cv", "constant_phi", 4, 4, "", None],
        ["k4cv", "psi_constant", 4, 4, "", None],
        ["k4cv", "double_diffraction", 4, 4, "", None],

        ["k6c", "bissector_horizontal", 6, 5, "delta", None],
        ["k6c", "bissector_vertical", 6, 4, "mu gamma", None],
        ["k6c", "constant_incidence", 6, 5, "mu", None],
        ["k6c", "constant_kphi_horizontal", 6, 4, "kphi delta", None],
        ["k6c", "constant_omega_vertical", 6, 4, "mu gamma", None],
        ["k6c", "constant_phi_horizontal", 6, 5, "delta", None],
        ["k6c", "constant_phi_vertical", 6, 4, "mu gamma", None],
        ["k6c", "double_diffraction_horizontal", 6, 5, "delta", None],
        ["k6c", "double_diffraction_vertical", 6, 4, "mu gamma", None],
        ["k6c", "lifting_detector_komega", 6, 3, "mu kappa kphi", None],
        ["k6c", "lifting_detector_kphi", 6, 3, "mu komega kappa", None],
        ["k6c", "lifting_detector_mu", 6, 3, "komega kappa kphi", None],
        ["k6c", "psi_constant_vertical", 6, 4, "mu gamma", None],
        # fmt: on
    ],
)
def test_engine_axes(dname, mode, n_r, n_w, e_a_c, c_a_c, e4cv, e4cv_renamed, e6c, k4cv, k6c, tardis):
    diffractometer_choices = dict(e4cv=e4cv, e4cv_renamed=e4cv_renamed, e6c=e6c, k4cv=k4cv, k6c=k6c, tardis=tardis)
    assert dname in diffractometer_choices

    dfrctmtr = diffractometer_choices[dname]
    calc = dfrctmtr.calc
    engine = calc.engine

    assert mode in engine.modes
    engine.mode = mode  # tests below apply to this engine and mode

    # engine: canonical names
    assert len(engine.axes_r) == n_r
    assert len(engine.axes_w) == n_w
    assert engine.axes_c == e_a_c.split()  # e_a_c : engine axes_c

    # calc: user names (might be renamed from canonical)
    assert len(calc.axes_r) == n_r
    assert len(calc.axes_w) == n_w
    assert calc.axes_c == (c_a_c or e_a_c).split()  # c_a_c : calc axes_c
